# API Reference

### Airship

Airship is a singleton with the following set of methods:

- [Airship](#airship)
  - [.configure(options)](#airshipconfigureoptions)
  - [.identify(entity)](#airshipidentifyentity)
  - [.flag(flagName)](#airshipflagflagname)
  - [.setErrorListener(listener)](#airshipseterrorlistenerlistener)
  - [.shutdown()](#airshipshutdown)

### Flag

Calling `Airship.flag` returns a `Flag` instance, which has following methods:

- [Flag](#flag)
  - [.getType()](#flaggettype)
  - [.isEnabled(entity)](#flagisenabledentity)
  - [.isEligible(entity)](#flagiseligibleentity)
  - [.getPayload(entity)](#flaggetpayloadentity)
  - [.getTreatment(entity)](#flaggettreatmententity)

---

## Airship methods

#### `Airship.configure(options)`

Configures the rules used for feature flagging.

Locally provided rules (aka **Airship Core**):

```javascript
Airship.configure({ flagConfig: {...} });
// => Promise {<resolved>: undefined}
```

The structure of the flagConfig object is documented [here](configuring-core.md)

Hosted rules (aka **Airship Cloud**):

```javascript
Airship.configure({envKey: 'onz2150xjon6pkjr'})
// => Promise {<pending>}
```

`envKey` is a key that you can obtain by signing up at [airshiphq.com](https://airshiphq.com)

`Airship.configure` returns a `Promise`. You will need to wait for the `Promise` to be resolved before being able to call `Airship.flag`:

```javascript
// using await keyword
await Airship.configure({envKey: 'onz2150xjon6pkjr'})
var entity = {id: 1}
var flag = Airship.flag('bitcoin-pay')
console.log(
  'Bitcoin payments are',
  flag.isEnabled(entity) ? 'enabled' : 'disabled'
)
// => Bitcoin payments are enabled
```

```javascript
// using Promise.then
Airship.configure({envKey: 'onz2150xjon6pkjr'}).then(function() {
  var entity = {id: 1}
  var flag = Airship.flag('bitcoin-pay')
  console.log(
    'Bitcoin payments are',
    flag.isEnabled(entity) ? 'enabled' : 'disabled'
  )
  // => Bitcoin payments are enabled
})
```

An option you can pass to Airship.configure is `subscribeToUpdates`, which defaults to `true`.

```javascript
await Airship.configure({envKey: 'onz2150xjon6pkjr', subscribeToUpdates: false})
```

Subscribing to updates makes the SDK listen for real-time updates to hosted rules, so that your application reflects the latest rules. In general, we recommend subscribing to updates on the backend, and not doing so on the frontend, so that your users do not see disappearing or appearing portions of the application on frontend re-renders.

#### `Airship.identify(entity)`

Identifies a default entity to be used when entity is omitted from flag methods. **Unavailable in NodeJS**

```javascript
var user = {id: 1}
Airship.identify(user)
var flag = Airship.flag('bitcoin-pay')

flag.isEnabled()
// => true
flag.isEnabled(user)
// => true
```

This is most commonly used to identify the signed on user in a front-end application, so that subsequent method calls to `flag` objects can omit the `entity` argument.

#### `Airship.publish(entities)`

Identifies an array of entities. **Available for Airship Cloud**

```javascript
var user1 = {id: 1}
var user2 = {id: 2}
Airship.publish([user1, user2])
// => Promise {<pending>}
```

This is most commonly used to batch identity entities to the hosted service. Usually, entity information is identified and sent to the hosted service lazily on a periodic basis, but this will do so immediately. Recommended usage is to batch publish all the users (or entities) for a specific flag before you deploy the flag in code so that you have a bird's-eye view of the audience upfront. This method is not needed for most workflows.

#### `Airship.flag(flagName)`

Returns a flag object associated with the given `flagName`.

```javascript
var flag = Airship.flag('bitcoin-pay')
flag.isEnabled({id: 1})
// => true
```

#### `Airship.setErrorListener(listener)`

Sets a function to receive error messages. By default error messages are output using `console.error`.

```javascript
var airshipErrorMessages = []
function airshipErrorListener(errorMessage) {
  airshipErrorMessages.push(errorMessage)
}

Airship.setErrorListener(airshipErrorListener)

// bad entity
Airship.identify({displayName: 'Joe'})

airshipErrorMessages
// => ["Each entity must have an `id` field"]
```

#### `Airship.shutdown()`

Cleans up resources allocated during `Airship.configure`.

```javascript
await Airship.shutdown()
```

## Flag methods

#### `flag.getType()`

Gets the type of flag.

```javascript
Airship.flag('bitcoin-pay').getType()
// => "basic"

Airship.flag('unknown-flag-with-no-rules').getType()
// => "uncategorized"
```

#### `flag.isEnabled(entity)`

Determines if flag is enabled for entity

```javascript
var flag = Airship.flag('bitcoin-pay')
flag.isEnabled({id: 1})
// => true
```

**To test a group entity make sure to set `isGroup: true` on the group entity object**

```javascript
var groupEntity = {id: 5, displayName: 'BitcoinExchangeUsers', isGroup: true}
var flag = Airship.flag('bitcoin-pay')
flag.isEnabled(groupEntity)
```

#### `flag.isEligible(entity)`

Determines if entity is within the targeted populations. **Unavailable in Airship Core**

```javascript
var flag = Airship.flag('bitcoin-pay')
flag.isEligible({id: 1})
// => true
```

**To test a group entity make sure to set `isGroup: true` on the group entity object**

```javascript
var groupEntity = {id: 5, displayName: 'BitcoinExchangeUsers', isGroup: true}
var flag = Airship.flag('bitcoin-pay')
flag.isEligible(groupEntity)
```

#### `flag.getPayload(entity)`

Retrieves the payload associated with the treatment assigned to the entity. **Unavailable in Airship Core**

```javascript
var flag = Airship.flag('bitcoin-pay')
flag.getPayload({id: 1})
// => {"bitcoin-price": "$9001"}
```

**To test a group entity make sure to set `isGroup: true` on the group entity object**

```javascript
var groupEntity = {id: 5, displayName: 'BitcoinExchangeUsers', isGroup: true}
var flag = Airship.flag('bitcoin-pay')
flag.getPayload(groupEntity)
```

#### `flag.getTreatment(entity)`

Retrieves the treatment assigned to the entity in a multivariate flag. **Unavailable in Airship Core**

```javascript
var flag = Airship.flag('color-theme')
flag.getTreatment({id: 1})
// => "purple"
```

**To test a group entity make sure to set `isGroup: true` on the group entity object**

```javascript
var groupEntity = {id: 9, displayName: 'LavenderLovers', isGroup: true}
var flag = Airship.flag('color-theme')
flag.getTreatment(groupEntity)
```
