import {setLogger} from './logger'
import AirshipCloud from './cloud'
import AirshipCore from './core'

const PLATFORM = 'nodejs'
const VERSION = '2.0.0'

const SDK_VERSION = `${PLATFORM}:${VERSION}`

const defaultEnv = new AirshipCore()
defaultEnv.configure({})

export class AirshipBase {
  constructor() {
    this.gatingInfoListeners = []
  }

  static _isDict(obj) {
    return obj !== undefined && obj !== null && obj.constructor === Object
  }

  async publish(objs) {
    if (this.environment) {
      await this.environment.publish(objs)
    } else {
      throw 'Airship must be configured first before `publish` can be called'
    }
  }

  // This will allow for async/await
  async configure(options) {
    if (!AirshipBase._isDict(options)) {
      throw '<options> must be dictionary'
    }

    const envKey = options.envKey
    const flagConfig = options.flagConfig

    if (!envKey && !flagConfig) {
      throw '<options> must contain envKey corresponding to an environment key or a flagConfig dictionary'
    }

    if (envKey) {
      if (this.environment) {
        await this.environment.shutdown()
      }
      this.environment = new AirshipCloud(
        this.handleGatingInfoUpdate.bind(this)
      )
      await this.environment.configure(envKey, options.subscribeToUpdates)
    } else {
      if (this.environment) {
        await this.environment.shutdown()
      }
      this.environment = new AirshipCore()
      await this.environment.configure(flagConfig)
    }
  }

  async shutdown() {
    if (this.environment) {
      await this.environment.shutdown()
      delete this.environment
    } else {
      throw 'Airship must be configured first before `shutdown` can be called'
    }
  }

  flag(flagName) {
    return (this.environment || defaultEnv).flag(flagName)
  }

  setErrorListener(fn) {
    setLogger(fn)
  }

  handleGatingInfoUpdate(gatingInfo) {
    this.gatingInfoListeners.forEach(listener => listener(gatingInfo))
  }

  addGatingInfoListener(listener) {
    this.gatingInfoListeners.push(listener)
  }

  removeGatingInfoListener(listener) {
    this.gatingInfoListeners = this.gatingInfoListeners.filter(
      l => l !== listener
    )
  }

  identify(obj) {
    if (this.environment) {
      this.environment.identify(obj)
      this.environment._identifyObject(obj)
      this.environment.maybeIngest(true)
    } else {
      throw 'Airship must be configured first before `identify` can be called'
    }
  }
}

const Airship = new AirshipBase()

export default Airship
